/* Author : Paul Ortyl <ortylp@3miasto.net>
 * License: GPL
 * Version: 0.24C
 * Date   : 07.12.2001
 * Description: see email2SMS.lsm
 * Note   : If you want to be notified of next release
 *           of the script mailto:ortylp@3miasto.net
 * Bug/wish reports : mailto:ortylp@3miasto.net
 */

/*
 TODO: 
 *  - add more documentation
 *  - split the code for a better readability
 *  - replace some regexp methods with simplier and faster ones
 *            (was caused by almost literal rewrite from PERL)
 *  - the code may be faster (some often used regexps can be precompiled)
 *  - avoid (if possible) unneded malloc/free pairs calls
 *  - and many more things...
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <string.h>
#include "email2SMS.h"
//#include "list.h"

tblptr * tableptrs;  //table of pointers to subtitution rules
int listfilelines=0;
struct sitem *headerfields=NULL;
unsigned int counter=0;
static char * output=NULL;                    /* buffer for output string */
unsigned int MULTILINE=0;               /* apply rules for already prepared output 
					 * string too (0=no; 1=yes)                */
unsigned int SHORT_MAIL_LEN=0;         /* if mail is not longer than 
					*  SHORT_MAIL_LEN then send it as a whole*/
unsigned int SMSLEN=130;               /* default length of output text in bytes */
unsigned int skipcontents=1;           /* skip mail contents 0=no; 1=yes          */
unsigned int showattachmentdescr=0;    /* show description of attachments         */
unsigned int showattachmentfilename=0; /* show filenames of attachments           */
unsigned int countsize=0;	       /* count size of the mail                  */
char * signature=NULL;                 /* string for signature detection 
					* default should be: signature="^--";     */
char * headersep=NULL;                 /* default separating header from contents */
struct sitem *boundary=NULL;
struct iitem *MIME=NULL;
unsigned int MIME_enriched=0;  /* MIME text/enriched (but not multipart) detected */  
char * listfile=NULL;
regex_t *regsignature=NULL;

int main(int argc, char **argv)
{
  char * table;
  char * headerf=NULL;
  char * header=NULL; // pointer to header text
  
  CLI_args(argc, argv);
  listfilelines=load_list_file(listfile, &tableptrs, &table);  
  free(listfile);
  
  //concatenate headerfields into geaderf
  if (headerfields) {
    headerf=malloc(strlen(headerfields->str)+1);
    strcpy(headerf,headerfields->str);
    sdel(&headerfields);
    while(headerfields) {
      int len=strlen(headerf);
      headerf=realloc(headerf,len+strlen(headerfields->str)+2);
      headerf[len]='|';
      strcpy(headerf+len+1,headerfields->str);
      sdel(&headerfields);
    }
  }
  
  read_header(&header);
    
  { //FIXME use precompiled regexps
    int headerlen;
    int headerlines=0;
    int r, line=1;
    char **listofheaderlines;
    if ((headerlen=strlen(header))<20) {exit(1);}
    
    headerlines=trc(header,'\n',0);//count_lines(header);
    listofheaderlines=malloc(sizeof(char*)*headerlines);
    listofheaderlines[0]=header;

    //merge splited header lines
    for(r=0; line<headerlines; r++) {
      if (!header[r]) {
	if ((header[r+1]=='\t')||(header[r+1]==' '))
	  {header[r+1]=header[r]=' '; headerlines--; }
	else
	  {listofheaderlines[line++]=header+r+1;}
      }
    }
    //analyse header
    for(r=0; r<headerlines; r++) { 
      if (!MIME) {
	if (!findreg1(listofheaderlines[r],"^Content-Type: multipart",
		      REG_ICASE|REG_EXTENDED|REG_NOSUB)) {
	  ipush(&MIME,extract_boundary_string(&boundary,
					      listofheaderlines[r]));
	}
      }
      if (!findreg1(listofheaderlines[r],"^Content-Type: text/enriched",
		    REG_ICASE|REG_EXTENDED|REG_NOSUB))
	{MIME_enriched=1;}
      if (!findreg1(listofheaderlines[r],headerf,
		    REG_ICASE|REG_EXTENDED|REG_NOSUB))	{
	char * buf;
	buf=malloc(strlen(listofheaderlines[r])+1);
	strcpy(buf,listofheaderlines[r]);
	isodecod(&buf);
	myprint(buf,tableptrs,listfilelines);
	free(buf);
      }
    }
    free(listofheaderlines);
    free(header);
  }

  if (headersep) myprint(headersep,tableptrs,listfilelines);

  PGP_check_init();
  
  if (MIME) {
    mimeprocessing();
    idel(&MIME);
  }
  else {
    char * bufor;
    bufor=malloc(MAXLINELEN);
    while((fgets(bufor,MAXLINELEN-1,stdin)))  {
      //if empty line => skip processing
      if (!bufor[1]) {counter++; continue;}
      //count size
      counter+=strlen(bufor);
      if (MIME_enriched)
	{ //FIXME
	  enriched_parsing(&bufor);
	}
      if (PGP_check(&bufor))
	{break;}
      if (signature) {
	if (!findreg1(bufor,signature,
		      REG_ICASE|REG_EXTENDED|REG_NOSUB)) {
	  while (fgets(bufor,MAXLINELEN-1,stdin)) 
	    {
	      counter+=strlen(bufor);
	    }
	}
      }
      myprint(bufor,tableptrs,listfilelines);
      if (MIME_enriched)
	{ //FIXME
	  // because of enriched_parsing(&bufor); above
	  free(bufor);bufor=malloc(MAXLINELEN);
	}
    }
    free(bufor);
  }

  //remove trailing spaces
  if (output)
    { 
      char siz[20];
      while(output[strlen(output)-1]==' ') 
	{output[strlen(output)-1]=0;}
      
      trcs(&output,' ');

      if (countsize)
	{
	if ((counter+51)/1024 > 10) 
	     { snprintf(siz,19,"%.0f",(counter+512)/1024.0);}
	else
	     {
		snprintf(siz,19,"%.1f",(counter+51)/1024.0);
		if (siz[strlen(siz)-1]=='0') {siz[strlen(siz)-2]=0;}
	     }
	  subst(&output, siz, 0,0);
	}


      if (strlen(output)>SHORT_MAIL_LEN)
	{ output[SMSLEN]=0;}

      fputs(output,stdout);
      free(output);
    }
  while(boundary) sdel(&boundary);

  free(headerf);
  if (listfilelines) {free(tableptrs);free(table);}
  if (regsignature) {free(regsignature);}
  return(0);
}

void usage()
{
    puts("email2SMS 0.23C\n"
	 "License: GPL\n"
	 "Author: Paul Ortyl <ortylp@from.pl>\n\n"
	 "Usage:\n"
	 "  -h, --help    prints this message\n"
	 "  -n number     number of bytes for the output string\n"
	 "  -N number     number of bytes for the output string\n"
	 "                   if the whole mail to be sent would be shorter\n"
	 "                   than this number\n"
	 "  -X string     string to be found in the mail header eg. \"^From: \"\n"
	 "                   (compare with 'formail' options)\n"
	 "  -FS           equivalent to:\n"
	 "                   -X \"^From: \" -X \"^Subject: \"\n"
	 "  -k            include contents of the mail\n"
	 "  -s string     string separating mail header from contents\n"
	 "  -ad           show attachment descriptions\n"
	 "  -af           show attachment filenames\n"
	 "  -c            show mail size in KB at the beginning of SMS\n"
	 "  -sig string   string (regexp) for detection of signatures\n"
	 "                    the most commonly used is ^--\n"
	 "  -M            apply all rules to already preprocessed string too\n"
	 "                   (all output treated as one line string)\n"
	 "  -l filename   file with substitution/removal rules\n"
	 "                file format:\n"
	 "                      orig|subst|comment\n"
	 "                one rule per line,\n"
	 "                'orig' is a regexp -- see `man grep` or `man perlre`\n"
	 "                   for more details\n\n"
	 "Example of .procmailrc file: \n"
	 "    :0c\n"
	 "    | ./email2SMS -n 135 -FS -ad -af -k -sig \"^--\" | mail \" +4860XXXXXXX@text.plusgsm.pl\"\n\n"
	 "Send bugs/comments/wishes to the author.\n");
    return;
}

void CLI_args(int argc, char** argv)
{
#define LISTFILE ".list"
  char * HOME;
  int r;
  {
    if ((HOME= getenv("HOME")))
      {
	int HOMElen=strlen(HOME);
	if (!(listfile=malloc(strlen(LISTFILE)+HOMElen+2)))
	  {perror("malloc error"); exit(1);};
	strcpy(listfile,HOME);
	listfile[HOMElen++]='/';
	strcpy(listfile+HOMElen,LISTFILE);
      }
  }
  
  if (argc==1) {usage();exit(0);}
  for(r=1;r<argc;r++)
    {
      if ((!strcmp(argv[r],"--help"))||(!strcmp(argv[r],"-h")))
	{ usage();exit(0);}
      if (!strcmp(argv[r],"-n")) {
	long int num=0;
	r++; 
	if (r>=argc) {
	  fprintf(stderr,"%s: a reasonable value needed after -n\n",argv[0]);
	  exit(1);
	}
	num = strtol(argv[r], NULL, 10);
	if ((num >1) && (num < (1<<16))) {
	  SMSLEN=num;
	  continue;
	}
	else {
	  fprintf(stderr,"%s: a reasonable value needed after -n\n",argv[0]);
	  exit(1);
	}
      }
      if (!strcmp(argv[r],"-X")) {
	r++;
      	if (r>=argc) {
	  fprintf(stderr,"%s: -X needs value\n",argv[0]);
	  exit(1);
	}
	spush(&headerfields,argv[r]);
	continue;
      }
      if (!strcmp(argv[r],"-k")) {
	skipcontents=0;
	continue;
      }
      if (!strcmp(argv[r],"-FS")) {
	spush(&headerfields,"^From: ");
	spush(&headerfields,"^Subject: ");
	continue;
      }
      if (!strcmp(argv[r],"-s")) {
	r++;
      	if (r>=argc) {
	  fprintf(stderr,"%s: -s needs value\n",argv[0]);
	  exit(1);
	}
	headersep=argv[r];
	continue;
      }
      if (!strcmp(argv[r],"-l")) {
	r++;
      	if (r>=argc) {
	  fprintf(stderr,"%s: -l needs value\n",argv[0]);
	  exit(1);
	}
	if (strlen(argv[r])>strlen(listfile))
	  {if (!(listfile=realloc(listfile, strlen(argv[r])+1)))
	     {perror("realloc error"); exit(1);};	  
	  };
	strcpy(listfile,argv[r]);
	continue;
      }
      if (!strcmp(argv[r],"-ad")) {
	showattachmentdescr=1;
	continue;
      }
      if (!strcmp(argv[r],"-af")) {
	showattachmentfilename=1;
	continue;
      }
      if (!strcmp(argv[r],"-c")) {
	countsize=1;
	continue;
      }
      if (!strcmp(argv[r],"-N")) {
	long int num=0;
	r++; 
	if (r>=argc) {
	  fprintf(stderr,"%s: a reasonable value needed after -N\n",argv[0]);
	  exit(1);
	}
	num = strtol(argv[r], NULL, 10);
	if ((num >1) && (num < (1<<16))) {
	  SHORT_MAIL_LEN=num;
	  continue;
	}
	else {
	  fprintf(stderr,"%s: a reasonable value needed after -N\n",argv[0]);
	  exit(1);
	}
      }
      if (!strcmp(argv[r],"-sig")) {
	r++;
      	if (r>=argc) {
	  fprintf(stderr,"%s: -sig needs value\n",argv[0]);
	  exit(1);
	}
	signature=argv[r];	
	regsignature=malloc(sizeof(regex_t));
	findreg_comp(regsignature,signature,REG_ICASE|REG_EXTENDED|REG_NOSUB);
	continue;
      }
      if (!strcmp(argv[r],"-M")) {
	MULTILINE=1;
	continue;
      }
    }
  if (!SHORT_MAIL_LEN) {SHORT_MAIL_LEN=SMSLEN;}
  if (SHORT_MAIL_LEN < SMSLEN) {
    unsigned int z;
    fprintf(stderr, "%s: Incorrect values -- let"
	    "'email2SMS -n len1 -N len2' then len1 < len2 \n",argv[0]);
    z=SMSLEN; 
    SMSLEN=SHORT_MAIL_LEN;
    SHORT_MAIL_LEN=z;
  }
}

void myprint(char* string, tblptr * tableptrs, int listfilelines)
{
  int r, oldlen,thislen;
  int ncont=1; /* ==1 if this line ends with '=\n' 
		  so no traling spaces allowed*/
  char * bufor;

  if ((thislen=strlen(string))<=1) return;
  bufor=malloc(strlen(string)+1);
  strcpy(bufor,string);

  if ((bufor[0]=='>') 
      ||
      ((bufor[0]=='=') && (bufor[1]=='3') && (bufor[2]=='E')))
    {
      thislen=strlen(bufor);
      while((bufor[thislen-2]=='=')&&(bufor[thislen-1]=='\n')) 
	{
	  free(bufor);
	  bufor=malloc(MAXLINELEN);
	  fgets(bufor,MAXLINELEN-1,stdin);
	  counter+=(thislen=strlen(bufor));
	}
      free(bufor);
      return;
    } 
  // else 

  if (bufor[thislen-1]=='\n')
    { 
      if (bufor[thislen-2]=='=') { bufor[thislen-2]=0;ncont=0;}
      else { bufor[thislen-1]=0;}
    }

  for(r=0; (r<listfilelines)/* &&(bufor[0])*/; r++) {
    sop(&bufor,&(tableptrs[r].reg), tableptrs[r].to);
  }
  
  isodecodtext(&bufor);
  trc(bufor,'\t',' ');
  trc(bufor,'\n',' '); //it happens after base64 decoding
  trcs(&bufor,' ');  
  if (!(thislen=strlen(bufor))) 
    { free(bufor);return;};
  
  if (ncont && thislen) 
    if (bufor[thislen-1]!=' ')
      {subst(&bufor," ",thislen,thislen);thislen++;}

  if (thislen) 
    {
      if (output) {oldlen=strlen(output); } else {oldlen=0;}
      output=realloc(output,oldlen+thislen+1);
      strcpy(output+oldlen,bufor);
    }
  free(bufor);

  if (MULTILINE) {
    for(r=0; r<listfilelines ; r++) {
      sop(&output,&(tableptrs[r].reg), tableptrs[r].to);
    }
  }

  if (strlen(output)>SHORT_MAIL_LEN+512)
    {
      int rr;
      bufor=malloc(MAXLINELEN);
      while((rr=fread(bufor,1,MAXLINELEN-1,stdin)))
	{counter+=rr;}
      free(bufor);
    }
}

int extract_boundary_string(struct sitem ** boundary, char * string)
{
  int b=0,e=0,r;
  char * bufor;
  if (!findreg(string,"boundary=[\"]?[!-\\}]*[\"]?[; ]?",&b,&e,
	       REG_ICASE|REG_EXTENDED))
    {
      bufor=malloc(strlen(string)+1);
      strncpy(bufor, string+b, e-b);
      e-=b; b=0;
      bufor[e]=0;
      //remove trailing spaces and '"'
      while((bufor[e-1]=='"')||(bufor[e-1]==' ')) 
	{e--; bufor[e]=0;}
      
      while(bufor[b++]!='='){}; // find '='
      if (bufor[b]=='"') b++;
      
      //escape some chars
      for(r=b;bufor[r];r++) {
	switch(bufor[r]) {
	case '+': subst(&bufor,"\\+",r,r+1);r++;break;
	case '(': subst(&bufor,"\\(",r,r+1);r++;break;
	case ')': subst(&bufor,"\\)",r,r+1);r++;break;
	default : break;						   
	}
      }
      spush(boundary,bufor+b);
      free(bufor);
      //check mutipart type, error must not happen
      findreg(string,"multipart/",&b,&e, REG_ICASE);
      
      if (!strncasecmp(string+e,"alternative",11))
	{
	  return(MULTIPART_ALTERNATIVE);
	}
      if (!strncasecmp(string+e,"encrypted",9))
	{
	  return(MULTIPART_ENCRYPTED);
	}

      /* only MULTIPART_ALTERNATIVE and MULTIPART_MIXED are processed 
       * properly, default is MULTIPART_MIXED, see rfc1341
       */
      /*
      if (strcasecmp(string+e,"digest"))
	{
	  return(MULTIPART_DIGEST);
	}
      if (strcasecmp(string+e,"paralel"))
	{
	  return(MULTIPART_PARALEL);
	}
      if (strcasecmp(string+e,"mixed"))
	{
	  return(MULTIPART_MIXED);
	}
       */

      return(MULTIPART_MIXED);
    }
  else return(0);
}

/*
 *  char **tableptrs needs to be freed
 *  char *table needs to be freed
 */
int load_list_file(char * listfile, tblptr ** tableptrs, char** table)
{
  int listfilelines=0;
  off_t list_file_size;
  FILE *f;
  
  if (!(f=fopen(listfile, "r"))) {
    fprintf(stderr, "email2SMS: Cannot open list file %s", listfile);
    *tableptrs=NULL;
    return(0);
  }
  
  { //get listfile size
    struct stat ss;
    if (stat(listfile,&ss))
      {perror("stat() error"); exit(1); }
    list_file_size=ss.st_size;
  }
  
  *table=calloc(1,list_file_size);
  if (fread(*table, list_file_size,1, f)!=1)
    { perror("fread error"); exit(1); }
  fclose(f);
  
  listfilelines=trc(*table,'\n', 0)+1;
  
  *tableptrs=malloc(sizeof(tblptr)*(listfilelines));
  
  { // create table of pointers to subtitution rules
    int r, //counter, table[r] points to currently processed char
      rcurr=0, 
      found=0,
      previous_r=0;
    
    for(r=0; found<listfilelines ; r++)
      {
	switch((*table)[r]) {
	case SEPARATOR : 
	  (*table)[r]=0;
	  if (!rcurr)
	    { 
	      if (strlen((*table)+previous_r))
		{
		  findreg_comp(&((*tableptrs)[found].reg),
			       (*table)+previous_r,REG_EXTENDED); 
		}
	      else
		{
		  //force break
		  listfilelines=found+1;
		}
	    }
	  if (rcurr==1) 
	    {
	      if (strlen((*table)+previous_r))
		{(*tableptrs)[found].to=(*table)+previous_r;}
	      else 
		{(*tableptrs)[found].to=NULL;}
	    }
	  rcurr++;
	  previous_r=r+1;
	  break;
	case 0 :
	  if (!rcurr) 
	    {listfilelines=found;break;}
	  if (rcurr==1) 
	    {
	      if (strlen((*table)+previous_r))
		{(*tableptrs)[found].to=(*table)+previous_r;}
	      else 
		{(*tableptrs)[found].to=NULL;}
	    }
	  rcurr=0;
	  previous_r=r+1;
	  found++;
	  break;
	}
      }  
  }
  return listfilelines;
}

void subst(char ** source, const char * sub, int b, int e)
{
  char * target;
  if (!(target=malloc(strlen(*source)-(e-b)+(sub ? strlen(sub):0)+1)))
    { perror("memalloc error"); exit(1); }
  if (b) {// inside
    if (!sub) {//delete
      strncpy(target,*source,b);
      strcpy(target+b,*source+e);
    } 
    else {//substitute
      strncpy(target,*source,b);
      strcpy(target+b,sub);
      strcpy(target+b+strlen(sub),*source+e);
    }
  } 
  else {// at the beginning
    if (!sub) { strcpy(target,*source+e);} //delete
    else {//substitute
      strcpy(target,sub);
      strcpy(target+strlen(sub),*source+e);
    }
  }
  free(*source); *source=target;
  return;
}


// perl 's///' statement
void s(char ** source, char *reg, char * to, int flags)
{
  int regerr;
  regmatch_t match[1];
  regex_t regcompiled;
  if ((regerr=regcomp(&regcompiled, reg, flags)))
    regexp_error(regerr,reg);
  if ((regerr=regexec(&regcompiled, *source , 1, match, 0)))
    {
      if (regerr==REG_NOMATCH) {return;}
      else { perror("regexec error\n"); exit(1);}
    }
  subst(source, to, match->rm_so, match->rm_eo);
  return;
}

// perl 's///o' statement
void so(char ** source, char *reg, char * to, int flags)
{
  int regerr;
  regmatch_t match[1];
  int so_old=0, eo_old=0, oldlen=0;
  regex_t regcompiled;
  if ((regerr=regcomp(&regcompiled, reg, flags)))
    regexp_error(regerr,reg);
  do 
    {
      if ((regerr=regexec(&regcompiled, *source , 1, match, 0)))
	{
	  if (regerr==REG_NOMATCH) {return;}
	  else {perror("regexec error\n"); exit(1);}
	}
      subst(source, to, match->rm_so, match->rm_eo);
      if ((so_old==match->rm_so)&&(eo_old==match->rm_eo)
	  &&(oldlen==strlen(*source))) {return;}
      so_old=match->rm_so; eo_old=match->rm_eo; 
      oldlen=strlen(*source);
    } while(1);
}

// perl 's///o' statement with precompiled regexps
void sop(char ** source, regex_t *regcompiled, char * to)
{
  int regerr;
  regmatch_t match[1];
  int so_old=0, eo_old=0, oldlen=0;
  do 
    {
      if ((regerr=regexec(regcompiled, *source , 1, match, 0)))
	{
	  if (regerr==REG_NOMATCH) {return;}
	  else {perror("regexec error\n"); exit(1);}
	}
      subst(source, to, match->rm_so, match->rm_eo);
      if ((so_old==match->rm_so)&&(eo_old==match->rm_eo)
	  &&(oldlen==strlen(*source))) {return;}
      so_old=match->rm_so; eo_old=match->rm_eo; 
      oldlen=strlen(*source);
    } while(1);
}


/* find char f and substitute it with t, returns number of
 * found and substituted chars
 *
 * int trc(char *source, char f, char t)
 * char *source   // pointer to null terminated string
 * char f         // char to be found and substituted
 * char t         // substitution char
 */
inline int trc(char *source, char f, char t)
{
  int r, found=0;
  for(r=0; source[r]; r++)
    { if (source[r]==f) {source[r]=t; found++;} }
  return(found);
}

inline int count_lines(char *source)
{
  int r, found=0;
  for(r=0; source[r]; r++) {
    if (source[r]=='\n') found++;
  }
  return(found);
}

inline void trcs(char **source, char c)
{
  char buf[]="[ ]{2,}";
  char buf1[]=" ";
  buf[1]=buf1[0]=c;
  so(source,buf,buf1,REG_EXTENDED);  
  return;
}

void isodecodtext(char ** source)
{
  char *reg="=[0-9A-F]{2}";
  int regerr;
  int curposition=0;
  char character=0;
  regmatch_t match[1];
  regex_t regcompiled;
  if (regcomp(&regcompiled, reg, REG_EXTENDED))
    {perror("regexpcom error\n"); exit(1);};
  do 
    {
      if ((regerr=regexec(&regcompiled, *source+curposition , 1, match, 0)))
	{
	  if (regerr==REG_NOMATCH) {return;}
	  else
	    {
	      perror("regexec error\n"); exit(1);
	    }
	}
      curposition+=match->rm_so;
      if ((character=tochr((*source)[curposition+1],
			   (*source)[curposition+2])))
      { 
	(*source)[curposition]=character;
	curposition++;
	subst(source, NULL, curposition, curposition+2);
      }
      else {curposition+=3;}
    } while(1);
}

void isodecod(char ** source)
{
  char *reg="=\\?[-0-9A-z]{5,12}\\?[qbQB]{1}\\?[ ->A-z]{1,170}\\?=";
  int regerr;
  regmatch_t match[1];
  regex_t regcompiled;
  if (regcomp(&regcompiled, reg, REG_EXTENDED))
    {perror("regexpcom error\n"); exit(1);};
  do 
    { if ((regerr=regexec(&regcompiled, *source , 1, match, 0)))
      {
	if (regerr==REG_NOMATCH) {return;}
	else {perror("regexec error\n"); exit(1);}
      }
    /*===============================================================*/
      {
	char * bufor;
	char * bufor1;
	int bb,ee;
	bufor=malloc(match->rm_eo-match->rm_so+1);
	strncpy(bufor,*source+match->rm_so,match->rm_eo-match->rm_so);
	bufor[match->rm_eo-match->rm_so]=0;
	if (findreg(bufor, "\\?[bB]{1}\\?", &bb, &ee, REG_EXTENDED))
	  {// this space and lenght is not correct according to RFC, but works
	    if (findreg(bufor, "\\?[ ->A-z]{1,170}\\?=", &bb, &ee, REG_EXTENDED))
	      {perror("isodecod regexec error\n"); exit(1);}
	    bufor1=malloc(ee-bb+1);
	    strncpy(bufor1,*source+match->rm_so+bb+1,ee-bb-2);
	    bufor1[ee-bb-3]=0;
	    isodecodtext(&bufor1);
	    trc(bufor1,'_',' ');
	  }
	else
	  {
	    if (findreg(bufor, "\\?[ ->A-z]{1,170}\\?=", &bb, &ee, REG_EXTENDED))
	      {perror("isodecod regexec error\n"); exit(1);}
	    bufor1=malloc(ee-bb+1);
	    strncpy(bufor1,*source+match->rm_so+bb+1,ee-bb-2);
	    bufor1[ee-bb-3]=0;
	    from64(&bufor1);
	  }
	subst(source, bufor1 , match->rm_so, match->rm_eo);
	free(bufor);
	free(bufor1);
      }
      /*===========================================================*/
    } while(1);
}

char tochr(char H, char L)
{
  if (H<='9') {H-='0';}
  else {
    if ((H>='a') && (H<='f')) {H=H-'a'+10;}
    else { if ((H>='A') && (H<='F')) {H=H-'A'+10;}}
  }
  if (L<='9') {L-='0';}
  else {
    if ((L>='a') && (L<='f')) {L=L-'a'+10;}
    else { if ((L>='A') && (L<='F')) {L=L-'A'+10;}}
  }
  return((char) ((H<<4)+L));
}

char tochr64(char H)
{
  if (H<='9') {
    switch(H) {
    case '+':H=62;break;
    case '/':H=63;break;
    default:H=H-'0'+52;
    }
  }
  else  {
    if ((H>='a') && (H<='z')) {H=H-'a'+26;}
    else { if ((H>='A') && (H<='Z')) {H=H-'A';}}
  }
  return(H);
}

void from64(char **string)
{
  int r,r1,len;
  char c1,c2,c3,c4;
  char * target;
  len=strlen(*string);
  if (!(target=malloc(((len*3)>>2)+3)))
    { perror("memalloc error"); exit(1); }
  
  for(r=0,r1=0;r<len;r+=4,r1+=3) {
    c1=tochr64(*(*string+r));
    c2=tochr64(*(*string+r+1));
    c3=(*(*string+r+2)=='=' ? 0 : tochr64(*(*string+r+2)));
    c4=(*(*string+r+3)=='=' ? 0 : tochr64(*(*string+r+3)));
    target[r1]=(c1<<2)|((c2&48)>>4);
    target[r1+1]=((c2&15)<<4)|(c3>>2);
    target[r1+2]=c4|(c3<<6);
    target[r1+3]=0;
  }
  free(*string); *string=target;
}

inline void findreg_comp(regex_t * regcompiled, char* reg, int flags)
{
  int regerr;
  if ((regerr=regcomp(regcompiled, reg, flags)))
    regexp_error(regerr,reg);
}

int findreg(char * source, char *reg, int *b, int *e , int flags )
{
  int regerr;
  regmatch_t match[1];
  regex_t regcompiled;
  if (!reg) {return(1);}
  if ((regerr=regcomp(&regcompiled, reg, flags)))
    regexp_error(regerr,reg);
  if ((regerr=regexec(&regcompiled, source , 1, match, 0))) {
    if (regerr==REG_NOMATCH) {return(1);}
    else { perror("regexec error\n"); exit(1);}
  }
  *b= match->rm_so ; *e=match->rm_eo;
  return(0);
}

int findregp(char * source, regex_t *regcompiled, int *b, int *e)
{
  int regerr;
  regmatch_t match[1];
  if ((regerr=regexec(regcompiled, source , 1, match, 0))) {
    if (regerr==REG_NOMATCH) {return(1);}
    else { perror("regexec error\n"); exit(1);}
  }
  *b= match->rm_so ; *e=match->rm_eo;
  return(0);
}

int findreg1(char * source, char *reg, int flags)
{
  int regerr;
  regex_t regcompiled;
  if (!reg) {return(1);}
  if (!source) {return(1);}
  if ((regerr=regcomp(&regcompiled, reg, flags|REG_NOSUB)))
    regexp_error(regerr,reg);
  if ((regerr=regexec(&regcompiled, source , 0, NULL, 0))) {
    if (regerr==REG_NOMATCH) {return(1);}
    else { perror("regexec error\n"); exit(1);}
  }
  return(0);
}

int findreg1p(char * source, regex_t *regcompiled)
{
  int regerr;
  if (!source) {return(1);}
  if ((regerr=regexec(regcompiled, source , 0, NULL, 0))) {
    if (regerr==REG_NOMATCH) {return(1);}
    else { perror("regexec error\n"); exit(1);}
  }
  return(0);
}


void mimeprocessing()
{
  char * bufor, *uboundary, *header;
  regex_t regboundary, regfinalboundary;
  int boundarylen=strlen(boundary->str);
  
  //create boundary "^--boundary$" and "^--boundary--$"
  /*  uboundary=malloc(boundarylen+7);
  strcpy(uboundary+3,boundary->str);
  uboundary[0]='^';
  uboundary[1]=uboundary[2]='-';
  uboundary[boundarylen+3]='$';
  uboundary[boundarylen+4]=0;

  findreg_comp(&regboundary,uboundary,REG_NOSUB|REG_EXTENDED);

  uboundary[boundarylen+3]=uboundary[boundarylen+4]='-';  
  uboundary[boundarylen+5]='$';
  uboundary[boundarylen+6]=0;

  findreg_comp(&regfinalboundary,uboundary,REG_NOSUB|REG_EXTENDED);
  */
  uboundary=malloc(boundarylen+7);
  strcpy(uboundary+2,boundary->str);
  uboundary[0]=uboundary[1]='-';
  uboundary[boundarylen+2]=0;

  findreg_comp(&regboundary,uboundary,REG_NOSUB|REG_EXTENDED);

  uboundary[boundarylen+2]=uboundary[boundarylen+3]='-';  
  uboundary[boundarylen+4]=0;

  findreg_comp(&regfinalboundary,uboundary,REG_NOSUB|REG_EXTENDED);

  free(uboundary);

  bufor=malloc(MAXLINELEN);
  //find firs boundary string
  while(fgets(bufor, MAXLINELEN-1, stdin)) { 
    counter+=strlen(bufor);
    if (!findreg1p(bufor,&regboundary))  break;
  }

  while (! feof(stdin))
    {
      //FIXME use precompiled regexps
      int headerlen;
      int BOUND=0; /* if boundary found BOUND=1 else BOUND=0 */
      if (!findreg1p(bufor,&regfinalboundary)) break;
      read_header(&header);

      trc(header,'\n','\t');      
      headerlen=strlen(header);
      
      //analyse header
      if (!findreg1(header,"Content-Type: multipart",
		    REG_ICASE|REG_EXTENDED|REG_NOSUB)) {
	ipush(&MIME,extract_boundary_string(&boundary,header));
	mimeprocessing(); //RECURRENCY
	idel(&MIME);
      }
      
      if (showattachmentdescr)
	{/* # [\x0A-\xff]* instead of .* because \t must not 
	    # be matched in this part of string
	 */ 
	  int e,b,d;
	  if (!findreg(header,"Content-Description: [ -\\}]*\t",&b,&e,
		       REG_ICASE|REG_EXTENDED))
	    {
	      char* buf;
	      d=e-b;
	      buf=malloc(d+1);
	      strncpy(buf,header+b,d);
	      buf[d]=0;
	      if ( d> 22) // if the description is not an empty string
		{ isodecod(&buf); myprint(buf,tableptrs,listfilelines);}
	      free(buf);
	    }
	}
      
      if (showattachmentfilename)
	{
	  int e,b,d;
	  if (!findreg(header,"(file)?name=\"[ -.#-z]*\"",&b,&e,
		       REG_ICASE|REG_EXTENDED)) 
	    {
	      //   version with '""'
	      char* buf;
	      d=e-b;
	      buf=malloc(d+1);
	      strncpy(buf,header+b,d);
	      buf[d]=0;
	      //	 if "name" instead of "filename"
	      if (!strncmp(buf,"name",4)) {subst(&buf,"file",0,0);}
	      isodecod(&buf);
	      s(&buf,"=\"","=",0);
	      trc(buf,'"',' '); 
	      myprint(buf,tableptrs,listfilelines);
	      free(buf);
	    }
	  else 
	    {
	      if (!findreg(header,"(file)?name=[ -.#-z]*\t",&b,&e,
			   REG_ICASE|REG_EXTENDED)) {  
		// version without '""'
		char * buf;
		d=e-b;
		buf=malloc(d+1);
		strncpy(buf,header+b,d);
		buf[d]=0;
		// if "name" instead of "filename"
		if (!strncmp(buf,"name",4)) {subst(&buf,"file",0,0);}
		isodecod(&buf);
		myprint(buf,tableptrs,listfilelines);
		free(buf);
	      }
	    }
	} //end showattachmentfilename
      
      if (!skipcontents) 
	{
	
	  if ((MIME->i==MULTIPART_ALTERNATIVE) || (MIME->i==MULTIPART_MIXED))
	    {
	      if ((!findreg1(header,"Content-Type: .*text\\/plain",
			     REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  && (findreg1(header,"Content-Type: .*name=",
			       REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  && (!findreg1(header,"Content-Transfer-Encoding: BASE64",
				REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  )
		{ // contents with base64 encoding 
		  // FIXME
		  // need to read whole attachment, then decode
		  // split into lines and then parse through 
		  // myprint()
		  
		  
		  if (signature) 
		    {
		      while((fgets(bufor,MAXLINELEN-1,stdin)))
			{
			  //if empty line => skip processing
			  if (!bufor[1]) {counter++; continue;}
			  //count size
			  counter+=strlen(bufor);
			  //should be before signature_check()
			  // or expect errors because of "^--" similarity string
			  if (!findreg1p(bufor,&regboundary)) 
			    {BOUND=1;break;}
			  
			  bufor[strlen(bufor)-1]=0;
			  from64(&bufor);
			  if (signature_check(bufor))
			    {break;}
			  myprint(bufor,tableptrs,listfilelines);
			  //because from64() has freed string and then
			  // allocated shorter, much shorter than 
			  // MAXLINELEN
			  free(bufor);bufor=malloc(MAXLINELEN);
			}
		    }
		  else // there is no signature detection
		    {
		      while((fgets(bufor,MAXLINELEN-1,stdin)))  
			{
			  //if empty line => skip processing
			  if (!bufor[1]) {counter++; continue;}
			  //count size
			  counter+=strlen(bufor);
			  // if $BOUNDARY found then  exit block
			  if (!findreg1p(bufor,&regboundary))
			    {BOUND=1;break;}
			  // else print $_;
			  bufor[strlen(bufor)-1]=0;
			  from64(&bufor);
			  myprint(bufor,tableptrs,listfilelines);
			  //because from64() has freed string and the 
			  // allocated shorter, much shorter than 
			  // MAXLINELEN
			  free(bufor);bufor=malloc(MAXLINELEN);
			}
		    }
		} // endif text/plain base64
	      else
	      if ((!findreg1(header,"Content-Type: .*text\\/plain",
			     REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  && (findreg1(header,"Content-Type: .*name=",
			       REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  && (findreg1(header,"Content-Transfer-Encoding: BASE64",
			       REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  )
		{ // endif text/plain NO base64 encoding
		  if (signature) 
		    {
		      while((fgets(bufor,MAXLINELEN-1,stdin)))
			{
			  //if empty line => skip processing
			  if (!bufor[1]) {counter++; continue;}
			  //count size
			  counter+=strlen(bufor);
			  if (!findreg1p(bufor,&regboundary)) 
			    {BOUND=1;break;}
			  if (PGP_check(&bufor))
			    {break;}
			  if (signature_check(bufor))
			    {break;}
			  myprint(bufor,tableptrs,listfilelines);
			}
		    }
		  else // there is no signature detection
		    {
		      while((fgets(bufor,MAXLINELEN-1,stdin)))  
			{
			  //if empty line => skip processing
			  if (!bufor[1]) {counter++; continue;}
			  //count size
			  counter+=strlen(bufor);
			  // if $BOUNDARY found then exit block
			  if (!findreg1p(bufor,&regboundary))
			    {BOUND=1;break;}
			  if (PGP_check(&bufor))
			    {break;}
			  // else print $_;
			  myprint(bufor,tableptrs,listfilelines);
			}
		    }
		} // endif text/plain
	      else
	      if ((!findreg1(header,"Content-Type: .*text\\/enriched",
			     REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  && (findreg1(header,"Content-Type: .*name=",
			       REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  && (findreg1(header,"Content-Transfer-Encoding: BASE64",
			       REG_ICASE|REG_EXTENDED|REG_NOSUB))
		  )
		{ 
		  if (signature) 
		    {
		      while((fgets(bufor,MAXLINELEN-1,stdin)))
			{
			  //if empty line => skip processing
			  if (!bufor[1]) {counter++; continue;}
			  //count size
			  counter+=strlen(bufor);
			  if (!findreg1p(bufor,&regboundary)) 
			    {BOUND=1;break;}
			  enriched_parsing(&bufor);
			  if (signature_check(bufor)) 
			    {break;}
			  myprint(bufor,tableptrs,listfilelines);		      
			  //because enriched_parsing() has freed string and then
			  // allocated shorter, much shorter than 
			  // MAXLINELEN
			  free(bufor);bufor=malloc(MAXLINELEN);
			  
			}
		    }
		  else // there is no signature detection
		    {
		      while((fgets(bufor,MAXLINELEN-1,stdin)))  
			{
			  //if empty line => skip processing
			  if (!bufor[1]) {counter++; continue;}
			  //count size
			  counter+=strlen(bufor);
			  if (!findreg1p(bufor,&regboundary))
			    {BOUND=1;break;}
			  enriched_parsing(&bufor);
			  // else print $_;
			  myprint(bufor,tableptrs,listfilelines);
			  //because enriched_parsing() has freed string and the 
			  // allocated shorter, much shorter than 
			  // MAXLINELEN
			  free(bufor);bufor=malloc(MAXLINELEN);
			}
		    }
		} // endif text/plain
	    } //end if MIME MULTIPART mixed alternative
	  if (MIME->i==MULTIPART_ENCRYPTED)
	    {
	      if (!findreg1(header,"Content-Type: .*application/octet-stream",
			    REG_ICASE|REG_EXTENDED|REG_NOSUB))
		{
		  while((fgets(bufor,MAXLINELEN-1,stdin)))  
		    {
		      //if empty line => skip processing
		      if (!bufor[1]) {counter++; continue;}
		      //count size
		      counter+=strlen(bufor);
		      // if $BOUNDARY found then  exit block
		      if (!findreg1p(bufor,&regboundary))
			{BOUND=1;break;}
		      if (PGP_check(&bufor))
			{break;}
		    }
		}
	      
	    }
	  if (!BOUND) //boundary had not been found yet
	    {
	      while (fgets(bufor,MAXLINELEN-1,stdin)) 
		{
		  //if empty line => skip processing
		  if (!bufor[1]) {counter++; continue;}
		  //count size
		  counter+=strlen(bufor);
		  if (!findreg1p(bufor,&regboundary)) break;
		} 
	    }
	} //end skip_contents
      free(header);
    } // end while (!eof(stdin))
  free(bufor);   
}


void inline regexp_error(int err,char *reg)
{
  fprintf(stderr,"err: %s\n%s\n",regex_errstr[err],reg);
  exit(1);
};

void read_header(char **header)
{// read in the header
  char * buf;
  buf=malloc(MAXLINELEN);
  *header=malloc(MAXLINELEN); (*header)[0]=0;
  while(fgets(buf, MAXLINELEN-1, stdin)) { 
    counter+=strlen(buf);
    if (buf[0]=='\n') break;
    *header=realloc(*header, strlen(*header)+strlen(buf)+1);
    strcat(*header,buf);
  }
  free(buf);
}

void enriched_parsing(char **bufor)
{
  so(bufor,"</?color>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"</?center>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"<nofill>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"</?flushright>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"<param>.*</param>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"<FontFamily>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"<smaller>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"</?italic>",NULL,REG_ICASE|REG_EXTENDED);
  so(bufor,"<bigger>",NULL,REG_ICASE|REG_EXTENDED);
}

/******************/
static regex_t pgpreg;
  
inline void PGP_check_init()
{
  findreg_comp(&pgpreg,"^-----BEGIN PGP MESSAGE-----\n$",
	       REG_ICASE|REG_EXTENDED|REG_NOSUB); 
  return;
}
inline int PGP_check(char **bufor)
{     
  if (findreg1p(*bufor,&pgpreg))
    return(0);
  {
    myprint("PGPmsg",tableptrs,listfilelines);
    return(1);
  }
}
/***********************/

inline int signature_check(char *bufor)
{     
  if (findreg1p(bufor,regsignature))
    return(0);
  else
    return(1);
}
